// *F* Developed 2013-2014 rejecht <rejecht at outlook dot com>

class TeamSummaryTicker extends Actor
	HideDropDown
	;

// Problem:
// - Anything spawned on a server gets replicated to all
// players if (RemoteRole > ROLE_None).

// Solution:
// - bOnlyRelevantToOwner
// -- Replication amount effectively halved for team specific data.

// Implementation Note 1:
// - A static array for each metric is used.
// - The metrics in the static arrays are connected using
// the index into the array.
// - Static arrays only send changed indices.
// - PRI typically gets sent once, unless a player
// or bot joins or leaves, in which case a shift occurs,
// but nothing compared to the original problem.

// Implementation Note 2:
// - The first pass updates and sorts team data.

// Implementation Note 3:
// - Normal replication rules apply; only values
// that have changed get replicated.

// #

const IndexTeam_Max = 2;
const IndexTeam_A = 0;
const IndexTeam_B = 1;
const IndexTeam_Unspecified = 255;

const NullVector = Vect (0.0f, 0.0f, 0.0f);

const TeamSizeLimit = 16;

var protected string TeamSummaryClassName;
var protected class<TeamSummary> TeamSummaryClass;

struct STeamSummary
{
	var RePRI RePRI[TeamSizeLimit];
	var Vector Position[TeamSizeLimit]; // Cached Location
	var int HitPoints[TeamSizeLimit]; // Pawn.Health + Pawn.ShieldStrength
	var int MaximumHitPoints[TeamSizeLimit]; // xP.SuperHealthMax + xP.ShieldStrengthMax
	var byte Flags[TeamSizeLimit];
};

var protected STeamSummary TeamSummaryArray[IndexTeam_Max];

// #Lifetime

event PostBeginPlay ()
{
	local Class<TeamSummary> LoadedTeamSummaryClass;


	Super.PostBeginPlay ();

	// Allow servers to use a custom handler.

	LoadedTeamSummaryClass = Class<TeamSummary>(DynamicLoadObject (TeamSummaryClassName, Class'Core.Class'));

	if (LoadedTeamSummaryClass != None)
	{
		TeamSummaryClass = LoadedTeamSummaryClass;

		// How often to update team data.

		if (Level.NetMode != NM_Standalone)
		{
			SetTimer (1.0f / TeamSummaryClass.Default.NetUpdateFrequency, True);
        }
        else
        {
			SetTimer (1.0f / 100.0f, True);
        }
	}
	else
	{
		Warn ("(LoadedTeamSummaryClass == None) - TeamSummaryClassName: <" $ TeamSummaryClassName $ ">");
    }
}

// #

event Timer ()
{
	local array<PlayerController> ReplicationTarget;
	local Controller Other;
	local RePRI OtherRePRI;
	local int
		TeamSummaryCount[IndexTeam_Max],
		IndexTeamSummary,
		IndexTeam_Other,
		Index;

	// Primary: Collect Team Data

	for (Other = Level.ControllerList; Other != None; Other = Other.NextController)
	{
		OtherRePRI = RePRI(Other.PlayerReplicationInfo);

		if
		(
			Other.bIsPlayer // Exclude Non-Player Controllers
			&&
			(OtherRePRI != None)
			&&
			!OtherRePRI.bOnlySpectator // Exclude Spectators
			&&
			(OtherRePRI.Team != None)
		)
		{
			IndexTeam_Other = OtherRePRI.Team.TeamIndex;
			IndexTeamSummary = TeamSummaryCount[IndexTeam_Other];

			if (IndexTeamSummary < TeamSizeLimit) // Grab the first #TeamSizeLimit# players.
			{
				TeamSummaryArray[IndexTeam_Other].RePRI[IndexTeamSummary] = OtherRePRI;

				UpdatePlayerMetrics (Other, OtherRePRI, TeamSummaryArray[IndexTeam_Other], IndexTeamSummary);

				++TeamSummaryCount[IndexTeam_Other];
            }

			// Secondary: Collect Replication Targets
			// - Only human players are valid replication targets.

			if (Other.IsA ('PlayerController'))
			{
				ReplicationTarget[ReplicationTarget.Length] = PlayerController(Other);
            }
		}
    }

	// Each player receives a tailor made list of teammates.

	for (Index = 0; Index < ReplicationTarget.Length; ++Index)
	{
		IndexTeam_Other = ReplicationTarget[Index].PlayerReplicationInfo.Team.TeamIndex;

		CopyTeamSummary
		(
			TeamSummaryArray[IndexTeam_Other],
			TeamSummaryCount[IndexTeam_Other],
			ReplicationTarget[Index]
		);
    }
}

protected function UpdatePlayerMetrics (Controller Other, RePRI OtherRePRI, out STeamSummary Out_TeamSummary, int IndexTeam)
{
	local xPawn xP;

	xP = xPawn(Other.Pawn);

	if (xP != None)
	{
		Out_TeamSummary.Position[IndexTeam] = xP.Location;
		Out_TeamSummary.HitPoints[IndexTeam] = xP.Health + xP.ShieldStrength;
		Out_TeamSummary.MaximumHitPoints[IndexTeam] = xP.SuperHealthMax + xP.ShieldStrengthMax;

		if (Other.Adrenaline < Other.AdrenalineMax)
		{
			Out_TeamSummary.Flags[IndexTeam] = 0;
        }
        else
        {
			Out_TeamSummary.Flags[IndexTeam] = 1;
        }
	}
}

// #

// - Each copy operation will run through all 16 entries.
// Description:
// - Clients get updates on their teammates.
// - Using out on the struct to avoid copying.
// - Excludes self from own team summary.

protected function CopyTeamSummary (out STeamSummary Out_ReadOnlyTeamSummary, int Count, PlayerController RePC)
{
	local RePRI OwnRePRI;
	local int Index;

	OwnRePRI = RePRI(RePC.PlayerReplicationInfo);

	for (Index = 0; Index < TeamSizeLimit; ++Index)
	{
		if
		(
			(Index < Count)
			&&
			(Out_ReadOnlyTeamSummary.RePRI[Index] != OwnRePRI)
		)
		{
			OwnRePRI.TeamSummary.RePRI[Index] = Out_ReadOnlyTeamSummary.RePRI[Index];
			OwnRePRI.TeamSummary.Position[Index] = Out_ReadOnlyTeamSummary.Position[Index];
			OwnRePRI.TeamSummary.HitPoints[Index] = Out_ReadOnlyTeamSummary.HitPoints[Index];
			OwnRePRI.TeamSummary.MaximumHitPoints[Index] = Out_ReadOnlyTeamSummary.MaximumHitPoints[Index];
			OwnRePRI.TeamSummary.Flags[Index] = Out_ReadOnlyTeamSummary.Flags[Index];
        }
        else
        {
			OwnRePRI.TeamSummary.RePRI[Index] = None; // The client will ignore the other metrics associated at this index when the PRI is None.
        }
	}
}

// #

DefaultProperties
{
	TeamSummaryClassName = "Re.TeamSummary"

	// Actor

	bHidden = True

	bReplicateMovement = False
	bSkipActorPropertyReplication = True

	RemoteRole = ROLE_None
}
